import json
from typing import TYPE_CHECKING

from metrics_layer.core.exceptions import (
    AccessDeniedOrDoesNotExistException,
    QueryError,
)

from .base import MetricsLayerBase
from .filter import Filter

if TYPE_CHECKING:
    from metrics_layer.core.model.project import Project


class DashboardLayouts:
    grid = "grid"


class DashboardElement(MetricsLayerBase):
    def __init__(self, definition: dict, dashboard, project) -> None:
        self.project: Project = project
        self.dashboard = dashboard
        self.validate(definition)
        super().__init__(definition)

    def get_model(self):
        model = self.project.get_model(self.model)
        if model is None:
            raise QueryError(
                f"Could not find model {self.model} referenced in dashboard {self.dashboard.name}."
            )
        return model

    def validate(self, definition: dict):
        required_keys = ["model"]
        for k in required_keys:
            if k not in definition:
                raise QueryError(f"Dashboard Element missing required key {k}")

    def to_dict(self):
        definition = json.loads(json.dumps(self._definition))
        definition["metrics"] = self.metrics
        definition["filters"] = self.parsed_filters(json_safe=True)
        return definition

    @property
    def slice_by(self):
        return self._definition.get("slice_by", [])

    @property
    def metrics(self):
        if "metric" in self._definition:
            metric_input = self._definition["metric"]
        else:
            metric_input = self._definition.get("metrics", [])
        return [metric_input] if isinstance(metric_input, str) else metric_input

    def _raw_filters(self):
        if self.filters is None:
            return []
        return self.filters

    def parsed_filters(self, json_safe=False):
        to_add = {"week_start_day": self.get_model().week_start_day, "timezone": self.project.timezone}
        return [f for raw in self._raw_filters() for f in Filter({**raw, **to_add}).filter_dict(json_safe)]

    def _error(self, element, error, extra: dict = {}):
        return self.dashboard._error(element, error, extra)

    def collect_errors(self):
        errors = []

        try:
            self.get_model()
        except (AccessDeniedOrDoesNotExistException, QueryError) as e:
            errors.append(
                self._error(self._definition.get("model"), str(e) + " in dashboard " + self.dashboard.name)
            )

        for field in self.metrics:
            if not self._function_executes(self.project.get_field, field):
                err_msg = f"Could not find field {field} referenced in dashboard {self.dashboard.name}"
                errors.append(self._error(self._definition.get("metrics"), err_msg))

        for field in self.slice_by:
            if not self._function_executes(self.project.get_field, field):
                err_msg = f"Could not find field {field} referenced in dashboard {self.dashboard.name}"
                errors.append(self._error(self._definition.get("slice_by"), err_msg))

        for f in self._raw_filters():
            if not self._function_executes(self.project.get_field, f.get("field")):
                err_msg = (
                    f"Could not find field {f.get('field')} referenced"
                    f" in a filter in dashboard {self.dashboard.name}"
                )
                errors.append(self._error(self._definition.get("filters"), err_msg))
        return errors

    @staticmethod
    def _function_executes(func, argument, **kwargs):
        try:
            func(argument, **kwargs)
            return True
        except Exception:
            return False


class Dashboard(MetricsLayerBase):
    def __init__(self, definition: dict, project) -> None:
        if "name" in definition:
            definition["name"] = self.normalize_name(definition["name"])

        if definition.get("layout") is None:
            definition["layout"] = DashboardLayouts.grid

        self.project: Project = project
        self.validate(definition)
        super().__init__(definition)

    @property
    def label(self):
        if self._definition.get("label"):
            return self._definition.get("label")
        return self.name.replace("_", " ").title()

    def validate(self, definition: dict):
        required_keys = ["name", "layout"]
        for k in required_keys:
            if k not in definition:
                raise QueryError(f"Dashboard missing required key {k}")

    def to_dict(self):
        definition = json.loads(json.dumps(self._definition))
        definition["elements"] = [e.to_dict() for e in self.elements()]
        definition["filters"] = self.parsed_filters(json_safe=True)
        return definition

    def _error(self, element, error, extra: dict = {}):
        line, column = self.line_col(element)
        return {
            **extra,
            "dashboard_name": self.name,
            "message": error,
            "line": line,
            "column": column,
            "reference_type": "dashboard",
            "reference_id": self.name,
        }

    def collect_errors(self):
        errors = []
        for f in self._raw_filters():
            try:
                self.project.get_field(f["field"])
            except Exception:
                err_msg = f"Could not find field {f['field']} referenced in a filter in dashboard {self.name}"
                errors.append(self._error(self._definition["filters"], err_msg))

        for element in self.elements():
            errors.extend(element.collect_errors())
        return errors

    def printable_attributes(self):
        to_print = ["name", "label", "description"]
        attributes = self.to_dict()
        attributes["type"] = "dashboard"
        return {key: attributes.get(key) for key in to_print if attributes.get(key) is not None}

    def _raw_filters(self):
        if self.filters is None:
            return []
        return self.filters

    def parsed_filters(self, json_safe=False):
        all_filters = []
        info = {"timezone": self.project.timezone}
        week_start_days = set(e.get_model().week_start_day for e in self.elements())
        if len(week_start_days) == 1:
            info["week_start_day"] = week_start_days.pop()
        else:
            info["week_start_day"] = None
        for f in self._raw_filters():
            clean_filters = Filter({**f, **info}).filter_dict(json_safe)
            for clean_filter in clean_filters:
                if clean_filter != {}:
                    all_filters.append(clean_filter)
        return all_filters

    def elements(self):
        elements = self._definition.get("elements", [])
        return [DashboardElement(e, dashboard=self, project=self.project) for e in elements]

    def _missing_filter_explore_error(self, filter_obj: dict):
        return (
            f"Argument 'explore' not found in the the filter {filter_obj} on dashboard "
            f"{self.name}. The 'explore' argument is required on filters for the whole dashboard."
        )



from typing import TYPE_CHECKING, List

import networkx as nx

from metrics_layer.core.exceptions import (
    AccessDeniedOrDoesNotExistException,
    JoinError,
    MetricsLayerException,
    QueryError,
)

from .base import MetricsLayerBase
from .field import Field
from .filter import Filter
from .join import Join, ZenlyticJoinRelationship, ZenlyticJoinType
from .view import View

if TYPE_CHECKING:
    from metrics_layer.core.model.project import Project


class Topic(MetricsLayerBase):
    valid_properties = [
        "version",
        "type",
        "name",
        "label",
        "model_name",
        "base_view",
        "description",
        "zoe_description",
        "hidden",
        "required_access_grants",
        "access_filters",
        "always_filter",
        "views",
        "extra",
    ]
    internal_properties = ["_file_path"]

    def __init__(self, definition: dict, project) -> None:
        self.project: Project = project
        self.validate(definition)
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["label", "base_view", "model_name"]
        for k in required_keys:
            if k not in definition:
                name_str = ""
                if k != "label":
                    name_str = f" in the topic {definition.get('label')}"
                raise QueryError(f"Topic missing required key {k}{name_str}")

    def id(self):
        return self.name

    @property
    def name(self):
        return self._definition.get("name", self.label)

    @property
    def model(self):
        try:
            return self.project.get_model(self.model_name)
        except AccessDeniedOrDoesNotExistException as e:
            e.message = str(e) + f" in topic {self.name}"
            raise e

    @property
    def hidden(self):
        try:
            model_is_hidden = self.model.hidden
        except Exception:
            model_is_hidden = False
        return bool(self._definition.get("hidden", False)) or model_is_hidden

    @property
    def base_view(self):
        return str(self._definition.get("base_view")).lower()

    def _views(self):
        topic_view_names = [self.base_view]
        if self.views and isinstance(self.views, dict):
            topic_view_names += list(self.views.keys())

        # Get all views - project now handles both regular views and from syntax
        topic_views = []
        for view_name in topic_view_names:
            try:
                view = self.project.get_view(view_name)
                topic_views.append(view)
            except AccessDeniedOrDoesNotExistException:
                pass

        return topic_views

    def from_view_references(self):
        if self.views and isinstance(self.views, dict):
            from_view_names = set(
                [v["from"] for k, v in self.views.items() if isinstance(v, dict) and "from" in v]
            )
            from_views = []
            for view_name in from_view_names:
                try:
                    view = self.project.get_view(view_name)
                    from_views.append(view)
                except AccessDeniedOrDoesNotExistException:
                    pass
            return from_views
        return []

    def _error(self, element, error, extra: dict = {}):
        line, column = self.line_col(element)
        return {
            **extra,
            "model_name": self.model_name,
            "message": error,
            "line": line,
            "column": column,
            "reference_type": "topic",
            "reference_id": self.label,
        }

    def always_filter_literal(self):
        to_add = {"week_start_day": self.model.week_start_day, "timezone": self.project.timezone}
        parsed_filters = []
        if self.always_filter:
            for f in self.always_filter:
                if "." not in f["field"]:
                    raise QueryError(
                        f"Always filter field {f['field']} in the topic {self.label} needs "
                        "to contain the view name like view_name.field_name."
                    )
                filter_dicts = Filter({**f, **to_add}).filter_dict(json_safe=False)
                for filter_dict in filter_dicts:
                    field_datatype = self.project.get_field(f["field"]).type
                    parsed_filters.append(
                        str(
                            Filter.sql_query(
                                f["field"],
                                filter_dict["expression"],
                                filter_dict["value"],
                                field_datatype,
                            )
                        )
                    )
        if parsed_filters:
            return " and ".join(parsed_filters)
        return None

    def collect_errors(self):
        errors = []

        if "name" in self._definition and not isinstance(self.name, str):
            errors.append(
                self._error(
                    self.name, f"The name property, {self.name} must be a string in the topic {self.label}"
                )
            )

        if "label" in self._definition and not isinstance(self.label, str):
            errors.append(
                self._error(
                    self.label, f"The label property, {self.label} must be a string in the model {self.name}"
                )
            )

        if "base_view" in self._definition:
            if not isinstance(self._definition["base_view"], str):
                errors.append(
                    self._error(
                        self._definition["base_view"],
                        (
                            f"The base_view property, {self._definition['base_view']} must be a string"
                            f" in the topic {self.label}"
                        ),
                    )
                )
            else:
                # Check if base_view exists
                try:
                    self.project.get_view(self.base_view)
                except AccessDeniedOrDoesNotExistException:
                    errors.append(
                        self._error(
                            self.base_view,
                            (
                                f"The base_view property, {self.base_view} in the topic {self.label} is not a"
                                " valid view"
                            ),
                        )
                    )

        if "model_name" in self._definition:
            if not isinstance(self.model_name, str):
                errors.append(
                    self._error(
                        self.model_name,
                        (
                            f"The model_name property, {self.model_name} must be a string in the topic"
                            f" {self.label}"
                        ),
                    )
                )
        else:
            try:
                self.model
            except AccessDeniedOrDoesNotExistException:
                errors.append(
                    self._error(
                        self.model_name,
                        (
                            f"The model_name property, {self.model_name} does not exist in the topic"
                            f" {self.label}"
                        ),
                    )
                )

        # This value is pulled from the VIEW_DESCRIPTION_MAX_CHARS constant in Zenlytic
        topic_description_max_chars = 2048
        if "description" in self._definition:
            if not isinstance(self.description, str):
                errors.append(
                    self._error(
                        self.description,
                        (
                            f"The description property, {self.description} must be a string in the topic"
                            f" {self.label}"
                        ),
                    )
                )
            else:
                if len(self.description) > topic_description_max_chars:
                    errors.append(
                        self._error(
                            self.description,
                            (
                                "Warning: The description property, must be"
                                f" {topic_description_max_chars} characters or less in the topic {self.label}"
                            ),
                        )
                    )
        if "zoe_description" in self._definition:
            if not isinstance(self.zoe_description, str):
                errors.append(
                    self._error(
                        self.zoe_description,
                        (
                            f"The zoe_description property, {self.zoe_description} must be a string in the"
                            f" topic {self.label}"
                        ),
                    )
                )
            else:
                if len(self.zoe_description) > topic_description_max_chars:
                    errors.append(
                        self._error(
                            self.zoe_description,
                            (
                                "Warning: The zoe_description property, must be"
                                f" {topic_description_max_chars} characters or less in the topic {self.label}"
                            ),
                        )
                    )
        if "hidden" in self._definition:
            if not isinstance(self._definition["hidden"], bool):
                errors.append(
                    self._error(
                        self._definition["hidden"],
                        (
                            f"The hidden property, {self._definition['hidden']} must be a boolean in the"
                            f" topic {self.label}"
                        ),
                    )
                )

        if "always_filter" in self._definition and not isinstance(self.always_filter, list):
            errors.append(
                self._error(
                    self._definition["always_filter"],
                    (
                        f"The always_filter property, {self.always_filter} must be a list in the topic"
                        f" {self.label}"
                    ),
                )
            )
        elif "always_filter" in self._definition and isinstance(self.always_filter, list):
            for f in self.always_filter:
                if not isinstance(f, dict):
                    errors.append(
                        self._error(
                            self._definition["always_filter"],
                            f"Always filter {f} in topic {self.label} must be a dictionary",
                        )
                    )
                    continue

                if "field" in f and isinstance(f["field"], str) and "." not in f["field"]:
                    errors.append(
                        self._error(
                            f["field"],
                            (
                                f"Always filter in topic {self.label} is referencing a field,"
                                f" {f['field']} that is not full qualified with view_name.field_name"
                            ),
                        )
                    )
                else:
                    errors.extend(
                        Field.collect_field_filter_errors(
                            f, self.project, "Always filter", "topic", self.label, error_func=self._error
                        )
                    )

        if "access_filters" in self._definition and not isinstance(self.access_filters, list):
            access_filter_error = self._error(
                self._definition["access_filters"],
                (
                    f"The topic {self.label} has an access filter, {self.access_filters} that is incorrectly"
                    " specified as a when it should be a list, to specify it correctly check the"
                    " documentation for access filters at"
                    " https://docs.zenlytic.com/docs/data_modeling/access_grants#access-filters"
                ),
            )
            errors.append(access_filter_error)
        elif self.access_filters is not None and isinstance(self.access_filters, list):
            for f in self.access_filters:
                if not isinstance(f, dict):
                    errors.append(
                        self._error(
                            self._definition["access_filters"],
                            f"Access filter {f} in topic {self.label} must be a dictionary",
                        )
                    )
                    continue
                if "field" not in f:
                    errors.append(
                        self._error(
                            self._definition["access_filters"],
                            f"Access filter in topic {self.label} is missing the required property: 'field'",
                        )
                    )
                elif "field" in f:
                    try:
                        self.project.get_field(f["field"])
                    except AccessDeniedOrDoesNotExistException:
                        errors.append(
                            self._error(
                                f["field"],
                                (
                                    f"Access filter in topic {self.label} is referencing a field,"
                                    f" {f['field']} that does not exist"
                                ),
                            )
                        )
                if "user_attribute" not in f:
                    errors.append(
                        self._error(
                            self._definition["access_filters"],
                            (
                                f"Access filter in topic {self.label} is missing the required user_attribute"
                                " property"
                            ),
                        )
                    )
                elif "user_attribute" in f and not isinstance(f["user_attribute"], str):
                    errors.append(
                        self._error(
                            f["user_attribute"],
                            (
                                f"Access filter in topic {self.label} is referencing a user_attribute,"
                                f" {f['user_attribute']} that must be a string, but is not"
                            ),
                        )
                    )

        errors.extend(
            View.collect_required_access_grant_errors(
                self._definition,
                self.project,
                f"in topic {self.label}",
                f"in model {self.model_name}",
                error_func=self._error,
            )
        )

        if "views" in self._definition and not isinstance(self.views, dict):
            errors.append(
                self._error(
                    self.views,
                    f"The views property, {self.views} must be a dictionary in the topic {self.label}",
                )
            )
        elif "views" in self._definition:
            for view_name, view_config in self.views.items():
                # Check if view exists
                try:
                    self.project.get_view(view_name)
                except AccessDeniedOrDoesNotExistException:
                    errors.append(
                        self._error(
                            view_name,
                            (
                                f"The view {view_name} in the views property of topic {self.label} does not"
                                " exist"
                            ),
                        )
                    )
                    continue

                if not isinstance(view_config, dict):
                    errors.append(
                        self._error(
                            view_config,
                            (
                                f"The view configuration for {view_name} in topic {self.label} must be a"
                                " dictionary"
                            ),
                        )
                    )
                    continue

                # Validate join configuration
                if "join" in view_config:
                    if not isinstance(view_config["join"], dict):
                        errors.append(
                            self._error(
                                view_config["join"],
                                (
                                    f"The join property for view {view_name} in topic {self.label} must be a"
                                    " dictionary"
                                ),
                            )
                        )
                        continue
                    join_config = view_config["join"]

                    # Check relationship property
                    valid_relationships = ZenlyticJoinRelationship.options
                    if "relationship" not in join_config:
                        errors.append(
                            self._error(
                                join_config,
                                (
                                    f"The join property for view {view_name} in topic {self.label} is missing"
                                    f" the required relationship property. Options are: {valid_relationships}"
                                ),
                            )
                        )
                    elif join_config["relationship"] not in valid_relationships:
                        errors.append(
                            self._error(
                                join_config["relationship"],
                                (
                                    f"The join property for view {view_name} in topic {self.label} has an"
                                    f" invalid relationship property. Options are: {valid_relationships}"
                                ),
                            )
                        )

                    # Check sql_on property
                    if "sql_on" not in join_config:
                        errors.append(
                            self._error(
                                join_config,
                                (
                                    f"The join property for view {view_name} in topic {self.label} is missing"
                                    " the required sql_on property"
                                ),
                            )
                        )
                    elif "sql_on" in join_config:
                        # Check for valid view references in sql_on
                        sql_on = join_config["sql_on"]
                        # Validate fields in sql_on
                        fields_to_replace = Join.fields_to_replace(sql_on)
                        for field in fields_to_replace:
                            view_name_part, column_name = Field.field_name_parts(field)
                            if view_name_part is None:
                                errors.append(
                                    self._error(
                                        sql_on,
                                        (
                                            f"Could not find view for field {field} in join for view"
                                            f" {view_name} in topic {self.label}"
                                        ),
                                    )
                                )
                                continue

                            try:
                                view = self.project.get_view(view_name_part)
                            except AccessDeniedOrDoesNotExistException:
                                err_msg = (
                                    f"Could not find view {view_name_part} in join for view {view_name} in"
                                    f" topic {self.label}"
                                )
                                errors.append(self._error(sql_on, err_msg))
                                continue

                            try:
                                self.project.get_field(column_name, view_name=view.name)
                            except AccessDeniedOrDoesNotExistException:
                                errors.append(
                                    self._error(
                                        sql_on,
                                        (
                                            f"Could not find field {column_name} in join for view"
                                            f" {view_name} in topic {self.label} referencing view"
                                            f" {view_name_part}"
                                        ),
                                    )
                                )

                    # Check join_type property
                    if "join_type" in join_config:
                        if join_config["join_type"] not in ZenlyticJoinType.options:
                            errors.append(
                                self._error(
                                    join_config["join_type"],
                                    (
                                        f"The join_type property for view {view_name} in topic"
                                        f" {self.label} must be one of {ZenlyticJoinType.options}"
                                    ),
                                )
                            )
                elif isinstance(self.base_view, str):
                    try:
                        # Check if the view can be joined to the base view automatically
                        self.project.join_graph.get_join(self.base_view, view_name)
                    except KeyError:
                        errors.append(
                            self._error(
                                view_name,
                                (
                                    f"The view {view_name} in topic {self.label} cannot be joined "
                                    f"automatically to the base view {self.base_view}. Please add an "
                                    "explicit join configuration or ensure the views share common "
                                    "identifiers that allow them to be joined."
                                ),
                            )
                        )
                # Validate 'from' property at view level
                if "from" in view_config:
                    from_view_name = view_config["from"]
                    if not isinstance(from_view_name, str):
                        errors.append(
                            self._error(
                                from_view_name,
                                (
                                    f"The from property for view {view_name} in topic {self.label} must be"
                                    " a string referencing an existing view name"
                                ),
                            )
                        )
                    else:
                        try:
                            self.project.get_view(from_view_name)
                        except AccessDeniedOrDoesNotExistException:
                            errors.append(
                                self._error(
                                    from_view_name,
                                    (
                                        f"The from property for view {view_name} references view"
                                        f" '{from_view_name}' which does not exist in topic {self.label}"
                                    ),
                                )
                            )

                if "override_access_filters" in view_config and not isinstance(
                    view_config["override_access_filters"], bool
                ):
                    errors.append(
                        self._error(
                            view_config["override_access_filters"],
                            (
                                f"The override_access_filters property for view {view_name} in topic"
                                f" {self.label} must be a boolean"
                            ),
                        )
                    )
                errors.extend(
                    self.invalid_property_error(
                        view_config,
                        {
                            "join",
                            "override_access_filters",
                            "from",
                            "label",
                            "field_prefix",
                            "include_metrics",
                        },
                        "topic",
                        self.label,
                        error_func=self._error,
                    )
                )

        if "extra" in self._definition and not isinstance(self.extra, dict):
            errors.append(
                self._error(
                    self.extra,
                    f"Topic {self.label} has an invalid extra {self.extra}. The extra must be a dictionary.",
                )
            )

        definition_to_check = {k: v for k, v in self._definition.items() if k not in self.internal_properties}
        errors.extend(
            self.invalid_property_error(
                definition_to_check, self.valid_properties, "topic", self.label, error_func=self._error
            )
        )
        return errors

    def printable_attributes(self):
        to_print = ["type", "label", "description"]
        attributes = self.to_dict()
        return {key: attributes.get(key) for key in to_print if attributes.get(key) is not None}

    def get_view_overrides(self, view_name: str) -> dict:
        if self.views and isinstance(self.views, dict):
            return self.views.get(view_name, {})
        return {}

    def query_validity_check(self, requested_views: List[str]) -> None:
        """
        Check if the requested views are valid for the topic.
        :param requested_views: list of view names that the user specifically asked for
        """
        if self.views and isinstance(self.views, dict):
            available_views = [self.base_view] + list(self.views.keys())
        else:
            available_views = [self.base_view]
        invalid_views = set(requested_views) - set(available_views)
        if invalid_views:
            invalid_views_str = ", ".join(sorted(invalid_views))
            raise JoinError(
                (
                    f"The following views are not included in the topic {self.label}:"
                    f" {invalid_views_str}\n\nYou can add them to the topic by adding the requested views to"
                    " the topic."
                ),
                location="topic",
            )

    def order_required_views(self, view_names: List[str]) -> List[str]:
        """Impute when other supporting joins are needed to complete a join
        :param view_names: list of view names that the user specifically asked for
        :return: list of view names in a valid topological order
        """
        self.query_validity_check(view_names)

        # If the topic has no joined views, return the view names
        if self.views is None and len(view_names) == 1:
            return view_names

        # Build a directed graph G where edges point from dependency -> dependent join
        G = nx.DiGraph()
        G.add_node(self.base_view)

        # Add edges for each join
        for view_name in self.views.keys():
            G.add_node(view_name)
            join = self.get_join(view_name)

            required_views = set(join.required_views()) - set([self.base_view, view_name])
            for required_view in required_views:
                G.add_edge(required_view, view_name)

        required_views = set()
        for r in view_names:
            required_views.add(r)
            required_views.update(nx.ancestors(G, r))  # all nodes that lead to r

        # Build a subgraph that only contains the required joins
        subG = G.subgraph(required_views).copy()

        sorted_views = list(nx.lexicographical_topological_sort(subG))
        return sorted_views

    def get_join(self, view_name: str) -> Join:
        # First check if the join exists in the topic as an override
        if view_name in self.views and "join" in self.views[view_name]:
            join_definition = {
                **self.views[view_name]["join"],
                "base_view_name": self.base_view,
                "join_view_name": view_name,
            }
            return Join(join_definition, self.project)
        try:
            join = self.project.join_graph.get_join(self.base_view, view_name)
            return join
        except KeyError:
            raise MetricsLayerException(
                f"Join not found between {self.base_view} and {view_name} in the topic {self.label}"
            )

    def join_graphs(self):
        join_graphs = []
        for view in self._views():
            join_graphs.append(self.project.join_graph.join_graph_hash(view.name))
        return join_graphs



import json
from collections import Counter
from typing import TYPE_CHECKING, Union

import pendulum

from metrics_layer.core.exceptions import (
    AccessDeniedOrDoesNotExistException,
    QueryError,
)

from .base import MetricsLayerBase
from .relationship import Relationship
from .view import View
from .week_start_day_types import WeekStartDayTypes

if TYPE_CHECKING:
    from metrics_layer.core.model.project import Project

SPECIAL_MAPPING_VALUES = {
    "date",
    "day_of_year",
    "week_of_year",
    "week",
    "month_of_year",
    "month",
    "quarter",
    "year",
}


class AccessGrant(MetricsLayerBase):
    valid_properties = ["name", "user_attribute", "allowed_values"]

    def __init__(self, definition: dict, model) -> None:
        self.model: Model = model
        self.validate(definition)
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["name", "user_attribute", "allowed_values"]
        for k in required_keys:
            if k not in definition:
                name_str = ""
                if k != "name":
                    name_str = f" {definition.get('name')}"
                raise QueryError(f"Access Grant{name_str} missing required key {k}")

    def id(self):
        return self.name

    def _error(self, element, error):
        return self.model._error(element, error)

    def collect_errors(self):
        errors = []
        if not self.valid_name(self.name):
            errors.append(self._error(self.name, self.name_error("Access Grant", self.name)))

        if not isinstance(self.user_attribute, str):
            errors.append(
                self._error(
                    self.user_attribute,
                    (
                        f"The user_attribute property, {self.user_attribute} must be a string in the Access"
                        f" Grant {self.name}"
                    ),
                )
            )

        if not isinstance(self.allowed_values, list):
            errors.append(
                self._error(
                    self.allowed_values,
                    (
                        f"The allowed_values property, {self.allowed_values} must be a list in the Access"
                        f" Grant {self.name}"
                    ),
                )
            )
        elif all([not isinstance(value, str) for value in self.allowed_values]):
            errors.append(
                self._error(
                    self.allowed_values,
                    (
                        "All values in the allowed_values property must be strings in the Access Grant"
                        f" {self.name}"
                    ),
                )
            )

        errors.extend(
            self.invalid_property_error(
                self._definition, self.valid_properties, "access grant", self.name, error_func=self._error
            )
        )
        return errors


class Model(MetricsLayerBase):
    valid_properties = [
        "version",
        "type",
        "name",
        "label",
        "hidden",
        "connection",
        "week_start_day",
        "timezone",
        "default_convert_tz",
        "fiscal_month_offset",
        "access_grants",
        "mappings",
        "required_access_grants",
        "relationships",
    ]
    internal_properties = ["_file_path"]

    def __init__(self, definition: dict, project) -> None:
        self.special_mapping_values = SPECIAL_MAPPING_VALUES
        self.project: Project = project
        self.validate(definition)
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["name", "connection"]
        for k in required_keys:
            if k not in definition:
                name_str = ""
                if k != "name":
                    name_str = f" in the model {definition.get('name')}"
                raise QueryError(f"Model missing required key {k}{name_str}")

    def id(self):
        return self.name

    @property
    def fiscal_month_offset(self):
        if "fiscal_month_offset" in self._definition:
            return self._definition["fiscal_month_offset"]
        return 0

    @property
    def hidden(self) -> bool:
        return bool(self._definition.get("hidden", False))

    @property
    def access_grants(self):
        if "access_grants" in self._definition:
            if not isinstance(self._definition["access_grants"], list):
                raise QueryError(
                    f"The access_grants property, {self._definition['access_grants']} must be a list"
                )
            elif not all([isinstance(grant, dict) for grant in self._definition["access_grants"]]):
                raise QueryError(f"All access_grants in the access_grants property must be dictionaries")
            return self._definition["access_grants"]
        return []

    @property
    def relationships(self):
        if "relationships" in self._definition:
            if not isinstance(self._definition["relationships"], list):
                raise QueryError(
                    f"The relationships property, {self._definition['relationships']} must be a list in "
                    f"the model {self.name}"
                )
            elif not all([isinstance(rel, dict) for rel in self._definition["relationships"]]):
                raise QueryError(
                    f"All relationships in the relationships property must be dictionaries in the model "
                    f"{self.name}"
                )
            return self._definition["relationships"]
        return []

    @property
    def mappings(self):
        mappings = json.loads(json.dumps(self._definition.get("mappings", {})))

        if not isinstance(mappings, dict):
            raise QueryError(f"The mappings property, {mappings} must be a dictionary")

        for date_mapping in self.special_mapping_values:
            if date_mapping in mappings:
                raise QueryError(
                    f"The mapping name {date_mapping} is a reserved name and cannot be used as a mapping name"
                )
            description = (
                f"The {date_mapping} associated with the metric or metrics you have "
                "in your query. When in doubt, use this to trend metrics over time."
            )
            all_canon_dates = [f.canon_date for f in self.project.fields()]
            unique_canon_dates = Counter(all_canon_dates).most_common()
            fields_mapped = []
            for d, _ in unique_canon_dates:
                canon_date_id = f"{d}_{date_mapping}"
                if d is not None and self.project.does_field_exist(canon_date_id):
                    fields_mapped.append(canon_date_id)
            # Includes all canon_dates in the project, sorted by number of occurrences in the project
            map_data = {"fields": fields_mapped, "group_label": "Dates", "description": description}
            mappings[date_mapping] = map_data
        return mappings

    def _error(self, element, error, extra: dict = {}):
        line, column = self.line_col(element)
        return {
            **extra,
            "model_name": self.name,
            "message": error,
            "line": line,
            "column": column,
            "reference_type": "model",
            "reference_id": self.name,
        }

    def collect_errors(self):
        errors = []
        if not self.valid_name(self.name):
            errors.append(self._error(self.name, self.name_error("model", self.name)))

        if not isinstance(self.connection, str):
            errors.append(
                self._error(
                    self.connection,
                    f"The connection property, {self.connection} must be a string in the model {self.name}",
                )
            )

        if "label" in self._definition and not isinstance(self.label, str):
            errors.append(
                self._error(
                    self.label, f"The label property, {self.label} must be a string in the model {self.name}"
                )
            )

        if "fiscal_month_offset" in self._definition:
            if not isinstance(self.fiscal_month_offset, int):
                errors.append(
                    self._error(
                        self.fiscal_month_offset,
                        (
                            f"The fiscal_month_offset property, {self.fiscal_month_offset} must be an integer"
                            f" in the model {self.name}"
                        ),
                    )
                )

        if "week_start_day" in self._definition:
            if str(self.week_start_day) not in WeekStartDayTypes.options:
                errors.append(
                    self._error(
                        self.week_start_day,
                        (
                            f"The week_start_day property, {self.week_start_day} must be one of"
                            f" {WeekStartDayTypes.options} in the model {self.name}"
                        ),
                    )
                )

        if "timezone" in self._definition:
            try:
                pendulum.timezone(str(self.timezone))
            except Exception:
                errors.append(
                    self._error(
                        self.timezone,
                        (
                            f"The timezone property, {self.timezone} must be a valid timezone in the model"
                            f" {self.name}. Valid timezones can be found at"
                            " https://en.wikipedia.org/wiki/List_of_tz_database_time_zones"
                        ),
                    )
                )

        if "default_convert_tz" in self._definition and not isinstance(self.default_convert_tz, bool):
            errors.append(
                self._error(
                    self.default_convert_tz,
                    (
                        f"The default_convert_tz property, {self.default_convert_tz} must be a boolean in the"
                        f" model {self.name}"
                    ),
                )
            )

        try:
            if self.access_grants:
                for access_grant in self.access_grants:
                    try:
                        grant = AccessGrant(access_grant, model=self)
                        errors.extend(grant.collect_errors())
                    except QueryError as e:
                        errors.append(self._error(access_grant, str(e) + f" in the model {self.name}"))
        except QueryError as e:
            errors.append(
                self._error(self._definition["access_grants"], str(e) + f" in the model {self.name}")
            )
        errors.extend(
            View.collect_required_access_grant_errors(
                self._definition,
                self.project,
                f"in model {self.name}",
                f"in model {self.name}",
                error_func=self._error,
            )
        )

        # Validate relationships
        try:
            if self.relationships:
                for relationship in self.relationships:
                    try:
                        rel = Relationship(relationship, model=self)
                        errors.extend(rel.collect_errors())
                    except QueryError as e:
                        errors.append(self._error(relationship, str(e) + f" in the model {self.name}"))
        except QueryError as e:
            errors.append(
                self._error(self._definition.get("relationships"), str(e) + f" in the model {self.name}")
            )

        valid_mappings_properties = ["fields", "group_label", "description", "link", "label"]
        try:
            mappings = self.mappings
            for mapping_name, mapped_values in mappings.items():
                mapping_element = self._definition.get("mappings", {}).get(mapping_name)
                if not isinstance(mapping_name, str):
                    errors.append(
                        self._error(
                            mapping_element,
                            f"The mapping name, {mapping_name} must be a string in the model {self.name}",
                        )
                    )
                if not self.valid_name(mapping_name):
                    errors.append(
                        self._error(
                            mapping_element,
                            (
                                f"The mapping name, {mapping_name} is invalid. Please reference the naming"
                                " conventions (only letters, numbers, or underscores) in the model"
                                f" {self.name}"
                            ),
                        )
                    )

                if not isinstance(mapped_values, dict):
                    errors.append(
                        self._error(
                            mapping_element,
                            (
                                f"The mapping value, {mapped_values} must be a dictionary in the model"
                                f" {self.name}"
                            ),
                        )
                    )
                if isinstance(mapped_values, dict):
                    if "link" in mapped_values and not isinstance(mapped_values["link"], str):
                        errors.append(
                            self._error(
                                mapping_element,
                                (
                                    f"The link property, {mapped_values['link']} must be a string"
                                    f" in the mapping {mapping_name} in the model {self.name}"
                                ),
                            )
                        )

                    if "group_label" in mapped_values and not isinstance(mapped_values["group_label"], str):
                        errors.append(
                            self._error(
                                mapping_element,
                                (
                                    f"The group_label property, {mapped_values['group_label']} must be a"
                                    f" string in the mapping {mapping_name} in the model {self.name}"
                                ),
                            )
                        )
                    if "description" in mapped_values and not isinstance(mapped_values["description"], str):
                        errors.append(
                            self._error(
                                mapping_element,
                                (
                                    f"The description property, {mapped_values['description']} must be a"
                                    f" string in the mapping {mapping_name} in the model {self.name}"
                                ),
                            )
                        )
                    if not isinstance(mapped_values.get("fields"), list):
                        errors.append(
                            self._error(
                                mapping_element,
                                (
                                    f"The fields property, {mapped_values['fields']} must be a list"
                                    f" in the mapping {mapping_name} in the model {self.name}"
                                ),
                            )
                        )
                    elif isinstance(mapped_values.get("fields"), list):
                        for field_id in mapped_values["fields"]:
                            try:
                                self.project.get_field(field_id)
                            except AccessDeniedOrDoesNotExistException as e:
                                errors.append(
                                    self._error(
                                        mapping_element,
                                        f"In the mapping {mapping_name} in the model {self.name}, the "
                                        + str(e),
                                        {"field_name": field_id},
                                    )
                                )

                    errors.extend(
                        self.invalid_property_error(
                            mapped_values,
                            valid_mappings_properties,
                            "mapping",
                            mapping_name,
                            error_func=self._error,
                        )
                    )
        except QueryError as e:
            if "Field" in str(e) and "missing required key" in str(e):
                raise QueryError(str(e) + f" in the model {self.name}")
            errors.append(
                self._error(self._definition.get("mappings"), str(e) + f" in the model {self.name}")
            )

        definition_to_check = {k: v for k, v in self._definition.items() if k not in self.internal_properties}
        errors.extend(
            self.invalid_property_error(
                definition_to_check, self.valid_properties, "model", self.name, error_func=self._error
            )
        )
        return errors

    def printable_attributes(self):
        to_print = ["name", "type", "label", "connection"]
        attributes = self.to_dict()
        return {key: attributes.get(key) for key in to_print if attributes.get(key) is not None}

    def get_relationships(self, view_name: Union[str, None] = None):
        """
        Get all relationships for this model, optionally filtered by view name.

        Parameters
        ----------
        view_name : Union[str, None]
            If provided, only return relationships where the view is referenced
            in either from_table or join_table. If None, return all relationships.

        Returns
        -------
        list[dict]
            A list of relationship definitions, each containing:
            - from_table: str
            - join_table: str
            - join_type: str (optional)
            - relationship: str
            - sql_on: str
        """
        if not self.relationships:
            return []

        if view_name is None:
            # Return all relationships
            return self.relationships

        # Filter relationships by view_name
        filtered_relationships = []
        for rel in self.relationships:
            if rel.get("from_table") == view_name or rel.get("join_table") == view_name:
                filtered_relationships.append(rel)

        return filtered_relationships

    def get_mappings(self, dimensions_only: bool = False):
        if self.mappings is None:
            return {}

        dimension_mapping = {}
        for mapping_name, mapped_values in self.mappings.items():
            special_mapping = mapping_name in self.special_mapping_values

            for mapped_from_field in mapped_values.get("fields", []):
                # If the user doesn't have access to a mapped field, we skip it but continue
                # to give them access to the original field which they *do* have access to
                try:
                    from_field = self.project.get_field(mapped_from_field)
                except AccessDeniedOrDoesNotExistException:
                    continue
                if dimensions_only and from_field.field_type == "measure":
                    continue

                from_join_hash = self.project.join_graph.join_graph_hash(from_field.view.name)
                # Handles the only allowed dimensions groups, the default ones for date, week, etc
                if from_field.field_type in {"dimension_group"} and not special_mapping:
                    raise QueryError(
                        "This mapping is invalid because it contains a dimension group. "
                        f"Mappings can only contain dimensions or measures. Mapping with {from_field.id()}"
                    )

                for mapped_to_field in mapped_values.get("fields", []):
                    if mapped_to_field != mapped_from_field:
                        try:
                            to_field = self.project.get_field(mapped_to_field)
                        except AccessDeniedOrDoesNotExistException:
                            continue
                        to_join_hash = self.project.join_graph.join_graph_hash(to_field.view.name)
                        if to_field.field_type != from_field.field_type:
                            raise QueryError(
                                f"This mapping is invalid because the mapped fields {mapped_from_field} "
                                f"and {mapped_to_field} are not the same type"
                            )
                        reference = {"field": mapped_to_field, "to_join_hash": to_join_hash}
                        # Create an object that contains the mapping from the field to the fields it
                        # maps to with all their metadata under the references array
                        if mapped_from_field in dimension_mapping:
                            dimension_mapping[mapped_from_field]["references"].append(reference)
                        else:
                            map_data = {
                                "references": [reference],
                                "field_type": from_field.field_type,
                                "from_join_hash": from_join_hash,
                                "is_canon_date_mapping": special_mapping,
                            }
                            dimension_mapping[mapped_from_field] = map_data

        return dimension_mapping
